/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/11/22.
//

#ifndef FGGL_GRID_HEX_HPP
#define FGGL_GRID_HEX_HPP

#include <array>
#include <vector>
#include <cmath>
#include <compare>

#include <fggl/math/fmath.hpp>

/**
 * Hexagonal Grid Implementation.
 * Based largely off Amit's incredible grid documentation.
 */
namespace fggl::grid {

	enum class HexDirPointy {
		RIGHT = 0, TOP_RIGHT = 1, TOP_LEFT = 2, LEFT = 3, BOTTOM_LEFT = 4, BOTTOM_RIGHT = 5
	};

	enum class HexDirFlat {
		BOTTOM_RIGHT = 0, TOP_RIGHT = 1, TOP = 2, TOP_LEFT = 3, BOTTOM_LEFT = 4, BOTTOM = 5
	};

	constexpr std::array< std::array<int, 2>, 6> HEX_DIRECTIONS {{
		{1, 0}, {1, -1}, {0, 1},
		{-1, 0}, {-1, 1}, {0, 1}
	}};

	constexpr std::array< std::array<int, 2>, 6> HEX_DIAGONALS {{
																	{2, -1}, {+1, -2}, {-1, -1},
																	{-2, 1}, {-1, 2}, {1, 1}
																}};

	template<typename T>
	struct HexPointT {

		constexpr HexPointT(T posQ, T posR) : m_pos({posQ, posR}) {}
		constexpr explicit HexPointT(const std::array<int, 2>& pos) : m_pos(pos[0], pos[1]) {}

		[[nodiscard]]
		constexpr auto q() const -> T {
			return m_pos[0];
		}

		[[nodiscard]]
		constexpr auto r() const -> T{
			return m_pos[1];
		}

		[[nodiscard]]
		constexpr auto s() const -> T {
			return -m_pos[0]-m_pos[1];
		}

		inline HexPointT neighbour(HexDirPointy dir) {
			return this + HexPointT<T>(HEX_DIRECTIONS.at((int)dir));
		}

		inline HexPointT neighbour(HexDirFlat dir) {
			return this + HexPointT<T>(HEX_DIAGONALS.at((int)dir));
		}

		HexPointT operator+(const HexPointT<T>& other) const {
			return { m_pos[0] + other.m_pos[0], m_pos[1] + other.m_pos[1] };
		}

		HexPointT operator-(const HexPointT<T>& other) const {
			return { m_pos[0] - other.m_pos[0], m_pos[1] - other.m_pos[1] };
		}

		bool operator==(const HexPointT<T>& other) const {
			return m_pos[0] == other.m_pos[0] && m_pos[1] == m_pos[1];
		}

		auto operator<=>(const HexPointT<T>& other) const = default;

		[[nodiscard]]
		auto distance(const HexPointT& other) const -> T {
			auto vec = *this - other;
			return (
				::abs(vec.q())
				+ ::abs(vec.q() - vec.r())
				+ ::abs(vec.r()) / 2
				);
		}

		[[nodiscard]]
		auto hexesInRange(int range) const -> std::vector<HexPointT<T>> {
			std::vector<HexPointT<T>> results;
			for ( auto q = -range; q <= range; ++q ) {
				auto stopCount = std::min(range, -q+range);
				for ( auto r = std::max(-range, -q-range); r <= stopCount; ++r ) {
					results.push_back( this + HexPointT<T>(q, r) );
				}
			}
			return results;
		}

		private:
			std::array<T, 2> m_pos;
	};

	using FloatHex = HexPointT<float>;
	using IntHex = HexPointT<int>;

	template<typename T>
	constexpr FloatHex hexLerp(const HexPointT<T>& start, const HexPointT<T>& end, float t) {
		return {
			 math::lerp(start.q(), end.q(), t),
			 math::lerp(start.r(), end.r(), t)
		};
	}

	constexpr IntHex round(const FloatHex& hex) {
		// see https://observablehq.com/@jrus/hexround for original JS implementation
		float xGrid = std::round( hex.r() );
		float yGrid = std::round( hex.q() );
		float x = hex.q() - xGrid;
		float y = hex.r() - yGrid;
		auto dx = std::round(x + 0.5F*y) * (float)(x*x >= y*y);
		auto dy = std::round(y + 0.5F*x) * (float)(x*x < y*y);
		return { (int)(xGrid + dx), (int)(yGrid + dy) };
	}

	std::vector<IntHex> lineTo(const IntHex& start, const IntHex& end);

} // namespace fggl::grid

#endif //FGGL_GRID_HEX_HPP
